#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ucontext.h>

/*
 * Given two addresses and a FILE*,
 * prints the data between these addresses in hex to the file.
 */
void print_between_addresses_to_file(unsigned long long int start,
	unsigned long long int end, FILE *fp)
{
	unsigned char* startP = (unsigned char *)start;
        unsigned char* endP = (unsigned char *)end;
        while (startP != endP)
        {
                fprintf(fp, "%02x", ((const unsigned char *) startP)[0]);
                startP++;
        }
	fprintf(fp, "\n");
}

/*
 * Given a hexadecimal number in string format,
 * returns a decimal version.
 */
unsigned long long int string_address_to_int(char address[])
{
	char formattedAddress[32];
        strcpy(formattedAddress, "0x");
        strcat(formattedAddress, address);
        return strtoull(formattedAddress, NULL, 0);	
}

/*
 * Given a line from /proc/<pid>/maps,
 * parses the beginning address into the result variable.
 */
void parse_start_address(char *result, char *str, int offset)
{
	int j = 0;
	for (int i = offset; i < strlen(str); i++)
        {
                if (str[i] == '-')
                {
			result[j] = '\0';
                        break;
                }
                result[j] = str[i];
		j++;
        }
}

/*
 * Given a line from /proc/<pid>/maps,
 * parses the ending address into the result variable.
 */
void parse_end_address(char *result, char *str, int offset)
{
	int begin_recording = 0;
	int j = 0;
        for (int i = offset; i < strlen(str); i++)
        {
		if (begin_recording)
		{
			if (str[i] == ' ')
			{
				result[j] = '\0';
				break;
			}
			result[j] = str[i];
			j++;
		}
                if (str[i] == '-')
                {
                        begin_recording = 1;
                }
        }
}

/*
 * Given a line from /proc/<pid>/maps,
 * parses the permissions column into the result variable.
 */
void parse_permissions(char *result, char *str, int offset)
{
        int begin_recording = 0;
        int j = 0;
        for (int i = offset; i < strlen(str); i++)
        {
                if (begin_recording)
                {
                        if (str[i] == ' ')
                        {
                                result[j] = '\0';
                                break;
                        }
                        result[j] = str[i];
                        j++;
                }
                if (str[i] == ' ')
                {
                        begin_recording = 1;
                }
        }
}

/*
 * Given a line from /proc/<pid>/maps,
 * parses the pathname column into the result variable.
 */
void parse_pathname(char *result, char *str, int offset)
{
        int spaces_seen = 0;
        int j = 0;
        for (int i = offset; i < strlen(str); i++)
        {
		// There are a minimum of five spaces before the pathname column.
                if (spaces_seen == 5)
                {
			if (str[i] == ' ')
			{
				continue;
			}
                        if (str[i] == '\0' || str[i] == '\n')
                        {
                                result[j] = '\0';
                                break;
                        }
                        result[j] = str[i];
                        j++;
                }
                if (str[i] == ' ')
                {
                        spaces_seen++;
                }
        }
}

/*
 * Given some pointer to address and the size of the object,
 * prints all the bytes in that length to the file specified by fp
 */
void print_bytes_to_file(const void *object, size_t size, FILE *fp)
{
        size_t i;

        for(i = 0; i < size; i++)
        {
               fprintf(fp, "%02x", ((const unsigned char *) object)[i]);
        }
}

/*
 * Handler for when the 12 signal is received. Creates a checkpoint image of
 * the running program. This can be used in conjunction with myrestart to
 * continue the program.
 */
void interrupt_handler(int signum)
{
	printf("Saving checkpoint image\n");
	// Open the ckpt file to write to
	FILE *fp;
	fp = fopen("myckpt", "w'");

	pid_t myPid;
	// The filename /proc/<pid>/maps
	char filename[100];
	char pid[15];
	
	// Get the PID and use it to create the pid filename /proc/<pid>/maps
	myPid = getpid();
	sprintf(pid, "%d", myPid);
	strcpy(filename, "/proc/");
	strcat(filename, pid);
	strcat(filename, "/maps");

	// Open /proc/<pid>/maps and read it into memory.
	int fd = open(filename, O_RDONLY);
	int readRet;
	char buf[1024];
	int ret = read(fd, buf, 1024);

	// Parse through the file and write all memory sections to the myckpt file
	int i = 0;
	while (i < 1024)
	{
		char startAddress[128];
		char endAddress[128];
		char permissionStr[10];
		char pathnameStr[128];
		parse_start_address(startAddress, buf, i);
		parse_end_address(endAddress, buf, i);
		parse_permissions(permissionStr, buf, i);
		parse_pathname(pathnameStr, buf, i);
		// Each memory section has a line containing the addresses
		// permissions, and pathname
		fprintf(fp, "#M|%s|%s|%s|%s\n", startAddress, endAddress, permissionStr, pathnameStr);

		unsigned long long int startNum = string_address_to_int(startAddress);
		unsigned long long int endNum = string_address_to_int(endAddress);
		// The line above is followed by a line of the binary values
		// (in hexadecimal) of the data between the addresses
		print_between_addresses_to_file(startNum, endNum, fp);
		// Then the current position is moved to the start of the next
		// line of the file
		while (buf[i] != '\n' && buf[i] != '\0')
		{
			i++;
		}
		if (buf[i] == '\0')
			break;
		i++;
		if (buf[i] == '\0')
		break;
	}
	close(fd);

	// Creates a temporary file. If this file is not found after getting the context,
	// no more steps will be performed. This allows the context to be restored
	// without performing unneeded steps.
	FILE *fpTmp = fopen("myckpt.tmp", "ab+");
	fclose(fpTmp);
        // Print context to file
	ucontext_t con;
	ucontext_t *mycontext = &con;
	// Save the registers
	getcontext(mycontext);
	// If the file was created (we are saving a checkpoint file), then save the
	// bytes of the context to myckpt.
	// Otherwise, we must be restoring from a checkpoint and this is skipped.
	if ( access( "myckpt.tmp", F_OK ) != -1)
	{
	fprintf(fp, "#R\n");
	print_bytes_to_file(mycontext, sizeof *mycontext, fp);
	fprintf(fp, "\n");
	// End print context to file;
	fclose(fp);
	// Remove the tmp file when we are done saving.
	remove("myckpt.tmp");
	}
}

/*
 * Global constructor to run before the main method of the program
 * that this is linked to.
 * Upon receiving the 12 signal, the program will save a checkpoint
 * image of itself.
 */
__attribute__((constructor))
void myconstructor() 
{
	signal(SIGUSR2, interrupt_handler);
}
