#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ucontext.h>
#include <stdlib.h>

// The name of the checkpoint file
char ckpt_file[1000];
// The name of the memory section file
char pid_file[100];
// The starting address of the original stack
unsigned long long int original_stack_start;
// The length of the original stack
unsigned long long int original_stack_length;
// The checkpoint context from the checkpoint file
ucontext_t *ckpt_context;
// The data parsed in for the context from the checkpoint file
unsigned char ckpt_data[2000];

/*
 * From str (/proc/<pid>/maps), parses the start address of the memory
 * section line currently set by offset into result
 */
void parse_start_address(char *result, char *str, int offset)
{
	int j = 0;
	for (int i = offset; i < strlen(str); i++)
        {
                if (str[i] == '-')
                {
			result[j] = '\0';
                        break;
                }
                result[j] = str[i];
		j++;
        }
}

/*
 * From str (/proc/<pid>/maps), parses the end address of the line currently
 * set by offset into result
 */
void parse_end_address(char *result, char *str, int offset)
{
	int begin_recording = 0;
	int j = 0;
        for (int i = offset; i < strlen(str); i++)
        {
		if (begin_recording)
		{
			if (str[i] == ' ')
			{
				result[j] = '\0';
				break;
			}
			result[j] = str[i];
			j++;
		}
                if (str[i] == '-')
                {
                        begin_recording = 1;
                }
        }
}

/*
 * Transforms a hex number in string format into a decimal int.
 */
unsigned long long int string_address_to_int(char address[])
{
	char formattedAddress[32];
        strcpy(formattedAddress, "0x");
        strcat(formattedAddress, address);
        return strtoull(formattedAddress, NULL, 0);	
}

/*
 * Parses a pathname from the str (/proc/<pid>/maps) into result
 */
void parse_pathname(char *result, char *str, int offset)
{
        int spaces_seen = 0;
        int j = 0;
        for (int i = offset; i < strlen(str); i++)
        {
                if (spaces_seen == 5)
                {
			if (str[i] == ' ')
			{
				continue;
			}
                        if (str[i] == '\0' || str[i] == '\n')
                        {
                                result[j] = '\0';
                                break;
                        }
                        result[j] = str[i];
                        j++;
                }
                if (str[i] == ' ')
                {
                        spaces_seen++;
                }
        }
}

/*
 * Extremely long function which
 * -Removes original stack of this program
 * -Parses memory sections from checkpoint file
 * -Maps memory for the memory sections and maps the data to this memory.
 * -Restores the context given in the checkpoint file
 */
void restore_memory()
{
	// Remove the original stack of this program in case
	// the checkpoint image will be mapping memory to this location.
	unsigned long long int *startP = &original_stack_start;
	size_t lengthP = (size_t)original_stack_length;
	munmap(startP, lengthP);

	// Preparing to restore the memory sections from the checkpoint file
	FILE *fp;
	fp = fopen(ckpt_file, "r");
	char c;
	c = fgetc(fp);
	// Parse through the file.
	while (c != EOF)
	{
		c = fgetc(fp);
		// If we have reached the register section, we have finished all
		// memory sections and we are done.
		if (c == 'R')
		{
			break;
		}
		// This c is the first pipe character
		c = fgetc(fp);
		// This c is the first character of the start address
		c = fgetc(fp);
		char startAddress[64];
		char endAddress[64];
		char permissions[10];
		char pathname[128];
		int k = 0;
		// Parse the start address
		while (c != '|')
		{
			startAddress[k] = c;
			c = fgetc(fp);
			k++;
		}
		startAddress[k] = '\0';
		c = fgetc(fp);
		k = 0;
		// Parse the end address
		while (c != '|')
		{
			endAddress[k] = c;
			c = fgetc(fp);
			k++;
		}
		endAddress[k] = '\0';
		c = fgetc(fp);
		k = 0;
		// Parse the permissions
		while (c != '|')
		{
			permissions[k] = c;
			c = fgetc(fp);
			k++;
		}
		permissions[k] = '\0';
		c = fgetc(fp);
		k = 0;
		// Parse the pathname
		while (c != '\n')
		{
			pathname[k] = c;
			c = fgetc(fp);
			k++;
		}
		pathname[k] = '\0';
		// Convert the parsed addresses into numbers	
		unsigned long long int startNum = strtoull(startAddress, NULL, 16);
		unsigned long long int endNum = strtoull(endAddress, NULL, 16);
		size_t length = (size_t)(endNum - startNum);
		unsigned long long int *startNumP = (unsigned long long int *)startNum;

		// Parse the permissions
		int r = 0;
		int w = 0;
		int x = 0;
		int prot = 0;
		if (permissions[0] == 'r')
			r = 1;
		if (permissions[1] == 'w')
			w = 1;
		if (permissions[2] == 'x')
			x = 1;

		// Determine the permission flags
		if (r || w || x)
		{
			if (r)
				prot = prot | PROT_READ;
			if (w)
				prot = prot | PROT_WRITE;
			if (x)
				prot = prot | PROT_EXEC;
		}
		else
		{
			prot = PROT_NONE;
		}

		// If the section we are looking at is for [vsyscall], this is a
		// static address we don't want to map.
		int vsysCompare = 0;
		// strcmp would cause a segmentation fault here.
		// Manually comparing solved this issue.
		if (pathname[0] == '[' && pathname[1] == 'v' && pathname[2] == 's' && pathname[3] == 'y' && pathname[4] == 's' && pathname[5] == 'c' && pathname[6] == 'a' && pathname[7] == 'l' && pathname[8] == 'l' && pathname[9] == ']')
			vsysCompare = 1;
		// If we know that this section is [vsyscall], move ahead to the next
		// memory section
		if (vsysCompare)
		{
			c = fgetc(fp);
			while (c != '\n')
				c = fgetc(fp);
			c = fgetc(fp);
			continue;
		}
		
		// Map the memory for this memory section
		// We use static permissions here and then use mprotect to actually
		// set them later. We need to be able to write the data before then.
		int* assignedAddress = mmap(startNumP, length, PROT_READ|PROT_WRITE,MAP_ANONYMOUS|MAP_PRIVATE|MAP_FIXED, -1, 0);
		// If mmap could not give us the FIXED address, the program cannot
		// be restored. We quit if so.
		if (assignedAddress != (int*) startNumP)
		{
			printf("ERROR- Did not get assigned address.\n");
			exit(1);
		}

		unsigned char* startCharP = (unsigned char *)assignedAddress;
		c = fgetc(fp);
		int z = 0;
		char c2;
		char both[3];
		// We map the data for the memory section to its address now.
		while (c != '\n')
		{
			c2 = fgetc(fp);
        	        both[0] = c;
			both[1] = c2;
                	both[2] = '\0';
                	long int num = strtol(both, NULL, 16);
                	unsigned char bin = num;
                	startCharP[z] = bin;

                	z++;
                	c = fgetc(fp);
		}
		// Set the actual permissions for this memory section that
		// were specified in the checkpoint file
		mprotect(startNumP,length, prot);
		c = fgetc(fp);
	}
	// Done mapping the memory sections
	fclose(fp);
	// Restoring the old registers. This is the last step.
	setcontext(ckpt_context);
}

/*
 * Initialze the name of the /proc/<pid>/maps file
 * global variable
 */
void initialize_pid_file()
{
	pid_t myPid;
	char pid[15];

	myPid = getpid();
	sprintf(pid, "%d", myPid);
	strcpy(pid_file, "/proc/");
        strcat(pid_file, pid);
        strcat(pid_file, "/maps");

}

/*
 * From the /proc/<pid>/maps file, parse this program's stack start address and
 * length.
 */
void initialize_stack_variables()
{
	int fd = open(pid_file, O_RDONLY);
	char buf[1024];
	// Read the file into memory
	int ret = read(fd, buf, 1024);
	char startAddress[128];
	char endAddress[128];
	char pathname[128];

	// Parse the file and look for the memory section labeled '[stack]'
	int i = 0;
	while (i < 1024)
	{
		parse_pathname(pathname, buf, i);
		if (strcmp(pathname,"[stack]") == 0)
		{
			parse_start_address(startAddress, buf, i);
			parse_end_address(endAddress, buf, i);
			break;
		}
		while (buf[i] != '\n' && buf[i] != '\0')
		{
			i++;
		}
		if (buf[i] == '\0')
			break;
		i++;
		if (buf[i] == '\0')
			break;
	}
	close(fd);
	unsigned long long int startNum = string_address_to_int(startAddress);
	unsigned long long int endNum = string_address_to_int(endAddress);
	// Calculate and store the stack length and starting address.
	original_stack_length = endNum - startNum;
	original_stack_start = startNum;
}

/*
 * From the checkpoint file, look for the bytes that store the register context,
 * and read them into the global variables ckpt_data and ckpt_context.
 */
void initialize_context()
{
	FILE *fp;
	fp = fopen(ckpt_file, "r");
	char c;
	int atRegister = 1;
	c = fgetc(fp);
	// Parse the file until we see '#R' which indicates we have reached
	// the section that stored the context.
	while (atRegister)
	{
		c = fgetc(fp);
		if (c == '#')
		{
			c = fgetc(fp);
			if (c == 'R')
			{
				atRegister = 0;
			}
		}
	}
	while (c != '\n')
	{
		c = fgetc(fp);
	}
	c = fgetc(fp);
	char c2;
	char both[3];
	unsigned int combined;
	int i = 0;
	// Read the context data into memory.
	while (c != '\n')
	{
		c2 = fgetc(fp);
		both[0] = c;
		both[1] = c2;
		both[2] = '\0';
		long int num = strtol(both, NULL, 16);
		unsigned char bin = num;
		ckpt_data[i] = bin;

		i++;
		c = fgetc(fp);	
	}
	// Set the ckpt_context to the data we parsed in
	ckpt_context = (ucontext_t*)&ckpt_data;
	fclose(fp);
}

/*
 * Set up all initial global variables.
 * This includes the name of the /proc/<pid>/maps file,
 * the original stack address and length, as well as
 * the register context from the checkpoint file.
 */
void setup()
{
	initialize_pid_file();
	initialize_stack_variables();
	initialize_context();
}

/*
 * Given a filename argument to this,
 * restore the checkpoint image in that file.
 */
int main(int argc, char *argv[])
{
	// If we didn't get just 1 argument, print usage and return.
	if (argc != 2)
	{
		printf("Usage is: ./myrestart [CHECKPOINT_IMAGE_FILE]\n");
		return 1;
	}
	// Save the filename to a global variable
	strcpy(ckpt_file, argv[1]);
	// Setup all needed variables so we don't have to do these computations later
	setup();
	// Map new memory for the stack.
	unsigned char *addr = (unsigned char *)0x5300000;
	mmap(addr, 4096,PROT_READ|PROT_WRITE|PROT_EXEC, MAP_FIXED|MAP_ANONYMOUS|MAP_PRIVATE|MAP_GROWSDOWN,-1,0);
	// Switch stack pointer from initial stack to new stack
	asm volatile ("mov %0,%%rsp;" : : "g" (addr) : "memory");
	// Immediately make new call to function which will then use call frame from
	// new stack.
	restore_memory();
}
